import shlex
import subprocess
import textwrap

from .. import common
from . import _rtorrent_common, rtorrent_socket

name = 'rtorrent_http'

username = 'fnark'
password = 'fnorkfnork'
# username = ''
# password = ''
httpport = '50302'


def get_http_home(tmp_path):
    return common.get_home(name, tmp_path) / 'nginx'


def get_http_pidfile_path(tmp_path):
    return get_http_home(tmp_path) / 'nginx.pid'


def run_http_server(tmp_path):
    http_home = get_http_home(tmp_path)
    http_home.mkdir(parents=True, exist_ok=True)
    nginx_configpath = http_home / 'nginx.conf'
    scgi_params_path = http_home / 'scgi_params'
    htpasswd_path = http_home / 'htpasswd'
    socketfile = rtorrent_socket.get_socketpath(tmp_path)

    if username:
        proc = subprocess.run(
            shlex.split(f'htpasswd -i -c "{htpasswd_path}" "{username}"'),
            text=True,
            input=password,
        )
        assert htpasswd_path.exists(), proc.stdout

        auth_config = textwrap.dedent(rf'''
            auth_basic "Restricted";
            auth_basic_user_file {htpasswd_path};
        ''').strip()

    else:
        auth_config = ''

    with open(nginx_configpath, 'w') as f:
        f.write(textwrap.dedent(rf'''
        error_log {http_home}/nginx.error.log;
        pid {get_http_pidfile_path(tmp_path)};

        events {{
            worker_connections 768;
        }}

        http {{
            server {{
                listen 127.0.0.1:{httpport};
                error_log {http_home}/nginx.error.log;
                access_log {http_home}/nginx.access.log;

                {auth_config}

                location /RPC2 {{
                   include {scgi_params_path};
                   scgi_pass unix:{socketfile};
                }}
            }}
        }}
        ''').strip())

    with open(scgi_params_path, 'w') as f:
        f.write(textwrap.dedent(r'''
        scgi_param  REQUEST_METHOD     $request_method;
        scgi_param  REQUEST_URI        $request_uri;
        scgi_param  QUERY_STRING       $query_string;
        scgi_param  CONTENT_TYPE       $content_type;

        scgi_param  DOCUMENT_URI       $document_uri;
        scgi_param  DOCUMENT_ROOT      $document_root;
        scgi_param  SCGI               1;
        scgi_param  SERVER_PROTOCOL    $server_protocol;
        scgi_param  REQUEST_SCHEME     $scheme;
        scgi_param  HTTPS              $https if_not_empty;

        scgi_param  REMOTE_ADDR        $remote_addr;
        scgi_param  REMOTE_PORT        $remote_port;
        scgi_param  SERVER_PORT        $server_port;
        scgi_param  SERVER_NAME        $server_name;
        ''').strip())

    subprocess.Popen(
        ('/usr/sbin/nginx', '-c', nginx_configpath),
        shell=False,
    )


def run(tmp_path):
    home = common.get_home(name, tmp_path)

    run_http_server(tmp_path)

    if username:
        server_url = f'http://{username}:{password}@localhost:{httpport}'
    else:
        server_url = f'http://localhost:{httpport}'

    return {
        'server_name': name,
        'server_start_cmd': (
            'rtorrent '
            '-o system.daemon.set=true '
            f'-d "{home}/downloads" '
            f'-s "{home}" '
            f'-o "import={_rtorrent_common.get_common_configpath(name, tmp_path)}" '
            f'-o "import={rtorrent_socket.get_socket_configpath(tmp_path)}" '
        ),
        'server_stop_cmd': (
            'bash -c "'
            f'kill $(cut -d + -f 2 "{home}/rtorrent.lock") '
            '; '
            f'kill $(cat "{get_http_pidfile_path(tmp_path)}") '
            '"'
        ),
        'server_url': server_url,
        'client_name': 'rtorrent',
    }



import pytest

from aiobtclientrpc import _errors


@pytest.mark.parametrize(
    argnames='a, b, exp_equal',
    argvalues=(
        (_errors.Error('foo'), _errors.Error('foo'), True),
        (_errors.Error('foo'), _errors.Error('bar'), False),
        (_errors.Error('foo'), _errors.RPCError('foo'), True),
        (_errors.Error('foo'), _errors.RPCError('bar'), False),
        (_errors.Error('foo'), ValueError('foo'), NotImplemented),
        (_errors.Error('foo'), ValueError('bar'), NotImplemented),
    ),
    ids=lambda v: repr(v),
)
def test_Error_equality(a, b, exp_equal):
    equal = a.__eq__(b)
    assert equal is exp_equal


@pytest.mark.parametrize(
    argnames='msg, info, exp_repr',
    argvalues=(
        ('foo', None, "RPCError('foo', info=None)"),
        ('foo', 'more info', "RPCError('foo', info='more info')"),
        ('foo', {'arbitrary': 'object'}, "RPCError('foo', info={'arbitrary': 'object'})"),
    ),
    ids=lambda v: repr(v),
)
def test_RPCError(msg, info, exp_repr):
    if info is not None:
        exception = _errors.RPCError(msg, info=info)
    else:
        exception = _errors.RPCError(msg)
    assert repr(exception) == exp_repr


@pytest.mark.parametrize(
    argnames='a, b, exp_equal',
    argvalues=(
        (_errors.RPCError('foo'), _errors.RPCError('foo'), True),
        (_errors.RPCError('foo'), _errors.RPCError('bar'), False),
        (_errors.RPCError('foo', info='my info'), _errors.RPCError('foo', info='my info'), True),
        (_errors.RPCError('foo', info='my info'), _errors.RPCError('foo', info='your info'), False),
        (_errors.RPCError('foo'), _errors.Error('foo'), NotImplemented),
        (_errors.RPCError('foo'), _errors.Error('bar'), NotImplemented),
    ),
    ids=lambda v: repr(v),
)
def test_RPCError_equality(a, b, exp_equal):
    equal = a.__eq__(b)
    assert equal is exp_equal


class TranslationTarget(Exception):
    def __init__(self, *posargs, **kwargs):
        args = []
        if posargs:
            args.append(', '.join(f'{arg!r}' for arg in posargs))
        if kwargs:
            args.append(', '.join(f'{k}={v!r}' for k, v in kwargs.items()))
        super().__init__(', '.join(args))

rpc_exception_map = {
    r'^The environment is perfectly safe$': TranslationTarget(r'RUN FOR YOUR LIVES!'),
    # Positional arguments
    r'^The (\w+) fell (\w+)$': (
        TranslationTarget,
        r'\1:', r'I fell \2!', r'Bugger \2!', 'OK!',
    ),
    # Keyword arguments
    r'^A (?P<what>\w+) hit the (?P<who>\w+)$': (
        TranslationTarget,
        ('who', r'\g<who>'), ('says', r'I was hit by a \g<what>!'), ('but', 'Just kidding.'),
    ),
}

@pytest.mark.parametrize(
    argnames='rpc_error, exp_return_value',
    argvalues=(
        # No references
        (_errors.RPCError('The environment is perfectly safe'), TranslationTarget('RUN FOR YOUR LIVES!')),
        (_errors.RPCError('The environment is toppled over'), _errors.RPCError('The environment is toppled over')),

        # Numbered references in positional arguments
        (_errors.RPCError('The front fell off'), TranslationTarget('front:', 'I fell off!', 'Bugger off!', 'OK!')),
        (_errors.RPCError('The bottom fell down'), TranslationTarget('bottom:', 'I fell down!', 'Bugger down!', 'OK!')),
        (_errors.RPCError('The front escalated quickly'), _errors.RPCError('The front escalated quickly')),

        # Named references in positional arguments
        (_errors.RPCError('A wave hit the ship'), TranslationTarget(who='ship', says='I was hit by a wave!', but='Just kidding.')),
        (_errors.RPCError('A whale hit the blimp'), TranslationTarget(who='blimp', says='I was hit by a whale!', but='Just kidding.')),
        (_errors.RPCError('A whale punched the blimp'), _errors.RPCError('A whale punched the blimp')),
    ),
    ids=lambda v: repr(v),
)
def test_RPCError_translate(rpc_error, exp_return_value):
    return_value = rpc_error.translate(rpc_exception_map)
    assert type(return_value) is type(exp_return_value)
    assert str(return_value) == str(exp_return_value)



from .. import common
from . import _rtorrent_common

name = 'rtorrent_socket'


def get_socketpath(tmp_path):
    return common.get_home(name, tmp_path) / 'rpc.socket'


def get_socket_configpath(tmp_path):
    configpath = common.get_home(name, tmp_path) / 'rtorrent.socket.rc'
    with open(configpath, 'w') as f:
        f.write(f'network.scgi.open_local = {get_socketpath(tmp_path)}\n')
        f.write(
            'schedule2 = scgi_permission,0,0,'
            + r'"execute.nothrow=chmod,\"g+w,o=\",'
            + str(get_socketpath(tmp_path))
            + '"\n'
        )
    return configpath


def run(tmp_path):
    homepath = common.get_home(name, tmp_path)
    return {
        'server_name': name,
        'server_start_cmd': (
            'rtorrent '
            '-o system.daemon.set=true '
            f'-d "{homepath}/downloads" '
            f'-s "{homepath}" '
            f'-o "import={_rtorrent_common.get_common_configpath(name, tmp_path)}" '
            f'-o "import={get_socket_configpath(tmp_path)}" '
        ),
        'server_stop_cmd': (
            'bash -c "'
            f'kill $(cut -d + -f 2 "{homepath}/rtorrent.lock")'
            '"'
        ),
        'server_url': f'file:///{get_socketpath(tmp_path)}',
        'client_name': 'rtorrent',
    }



from aiobtclientrpc import __project_name__
project = __project_name__

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx_rtd_theme',
]

html_theme = "sphinx_rtd_theme"
html_theme_options = {
    'titles_only': True,
}

autosummary_generate = True
autosummary_imported_members = True

html_show_sourcelink = False  # Don't show links to rST code

templates_path = ['_templates']

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}



import base64
import glob
import os
import re


def get_home(name, tmp_path):
    homepath = tmp_path / f'{name}.home'
    homepath.mkdir(parents=True, exist_ok=True)
    return homepath


def get_test_torrents(map='filepath:infohash'):
    torrent_files = {}
    torrents_dirpath = os.path.join(os.path.dirname(__file__), 'torrents')
    for filepath in glob.glob(f'{torrents_dirpath}/*.torrent'):
        infohash = re.sub(r'^.*([a-zA-Z0-9]{40})\.torrent$', r'\1', filepath)
        if map == 'filepath:infohash':
            torrent_files[filepath] = infohash
        elif map == 'infohash:filepath':
            torrent_files[infohash] = filepath
        else:
            raise ValueError('Invalid map: {map}')
    assert torrent_files, torrent_files
    return torrent_files


def read_torrent_file(filepath):
    with open(filepath, 'rb') as f:
        filecontent = f.read()
    return str(base64.b64encode(filecontent), encoding='ascii')



import os

from .. import common
from . import base


class API(base.APIBase):
    def __init__(self, client):
        self.client = client

    async def perform_simple_request(self):
        result = await self.client.call('core.get_config')
        self.log('daemon.get_config:', result)
        assert result['dht'] is False
        assert result['utpex'] is False

    async def get_torrent_list(self):
        result = await self.client.call('core.get_torrents_status', filter_dict={}, keys=['name'])
        self.log('core.get_torrents_status:', result)
        infohashes = sorted(result.keys())
        return infohashes

    async def add_torrents(self, torrent_filepaths, download_path='/tmp/some/path', paused=True):
        # Register event handler for added torrents
        torrents_added = []

        def on_torrent_added(infohash, from_state_):
            self.log('torrent added:', infohash, from_state_)
            torrents_added.append(infohash)

        await self.client.add_event_handler('TorrentAddedEvent', on_torrent_added)

        # Add torrents (as_file is ignored because Deluge doesn't accept file
        # paths)
        for filepath in torrent_filepaths:
            result = await self.client.call(
                'core.add_torrent_file',
                filename=os.path.basename(filepath),
                filedump=common.read_torrent_file(filepath),
                options={
                    'add_paused': paused,
                    'save_path': download_path,
                },
            )
            self.log('core.add_torrent_file_async', result)

        assert torrents_added

        # Verify torrents were correctly added
        result = await self.client.call(
            'core.get_torrents_status',
            filter_dict={'id': torrents_added},
            keys=['paused', 'save_path'],
        )
        self.log('core.get_torrents_status:', torrents_added, result)
        for infohash, torrent in result.items():
            assert torrent['save_path'] == download_path, torrent['save_path']
            if paused:
                assert torrent['paused'] is True, repr(torrent['paused'])
            else:
                assert torrent['paused'] is False, repr(torrent['paused'])

        return sorted(torrents_added)

    async def remove_torrents(self, infohashes):
        # Register event handler
        torrents_removed = []

        def on_torrent_removed(infohash):
            self.log('torrent removed:', infohash)
            torrents_removed.append(infohash)

        await self.client.add_event_handler('TorrentRemovedEvent', on_torrent_removed)

        # Remove torrents
        for infohash in infohashes:
            result = await self.client.call(
                'core.remove_torrent',
                torrent_id=infohash,
                remove_data=False,
            )
            self.log(f'core.remove_torrent {infohash}', result)

        # Check if events occured as expected
        assert torrents_removed == infohashes

    async def on_torrent_added(self, handler):
        def handler_wrapper(infohash, *args, **kwargs):
            handler(infohash)
        await self.client.add_event_handler('TorrentAddedEvent', handler_wrapper)

    async def wait_for_torrent_added(self):
        await self.client.wait_for_event('TorrentAddedEvent')



from . import base


class API(base.APIBase):
    def __init__(self, client):
        self.client = client

    async def perform_simple_request(self):
        result = await self.client.call('system.listMethods')
        self.log('system.listMethods:', result)
        assert 'system.listMethods' in result
        assert 'system.multicall' in result
        assert 'load.normal' in result

        result = await self.client.call('dht.statistics')
        self.log('dht.statistics:', result)
        assert result['dht'] == 'off'

        result = await self.client.call('protocol.pex')
        self.log('protocol.pex:', result)
        assert result == 0

    async def get_torrent_list(self):
        result = await self.client.call('download_list', '')
        self.log('download_list:', result)
        infohashes = sorted(infohash.lower() for infohash in result)
        return infohashes

    STATE_PAUSED = 0
    STATE_STARTED = 1

    async def add_torrents(self, torrent_filepaths, download_path='/tmp/some/path', paused=False):
        # Add torrents
        calls = []
        method_name = 'load.raw' if paused else 'load.raw_start'
        for filepath in torrent_filepaths:
            calls.append({
                'methodName': method_name,
                'params': [
                    '',
                    open(filepath, 'rb').read(),
                    # Set download location
                    f'd.directory_base.set="{download_path}"',
                    # Untie torrent from .torrent file so rtorrent doesn't
                    # delete it when the torrent is removed.
                    'd.tied_to_file.set='
                ],
            })
        result = await self.client.call('system.multicall', calls)
        self.log('load:', result)

        # Get added torrent hashes from server
        infohashes = await self.client.call('download_list', '')
        infohashes = [infohash.lower() for infohash in infohashes]

        # Verify torrents where correctly added
        fields = ('hash', 'state', 'directory_base')
        multicall_params = (
            ['main']
            + [f'd.{field}=' for field in fields]
        )
        result = await self.client.call('d.multicall2', '', multicall_params)
        torrents = [{field: value for field, value in zip(fields, item)}
                    for item in result]
        self.log('torrents:', torrents)
        for torrent in torrents:
            location = torrent['directory_base']
            state = torrent['state']
            assert location == download_path, f'{location!r} != {download_path!r}'
            if paused:
                assert state == self.STATE_PAUSED, f'{state!r} != {self.STATE_PAUSED!r}'
            else:
                assert state == self.STATE_STARTED, f'{state!r} != {self.STATE_STARTED!r}'

        added_infohashes = sorted(torrent['hash'].lower() for torrent in torrents)
        self.log('added torrents:', added_infohashes)
        return added_infohashes

    async def remove_torrents(self, infohashes):
        for infohash in infohashes:
            result = await self.client.call('d.erase', infohash)
            self.log(f'd.erase {infohash}:', result)

    async def on_torrent_added(self, handler):
        # Raise NotImplementedError
        await self.client.add_event_handler('<event name>', handler)

    async def wait_for_torrent_added(self):
        # Raise NotImplementedError
        await self.client.wait_for_event('<event name>')



import os
import textwrap

from .. import common

name = 'deluge'

username = 'fnark'
password = 'fnorkfnork'
rpcport = '50100'


def get_configpath(home):
    configdir = home / 'deluged'
    os.makedirs(configdir, exist_ok=True)

    with open(configdir / 'auth', 'w') as f:
        f.write(f'{username}:{password}:10\n')

    with open(configdir / 'core.conf', 'w') as f:
        f.write(textwrap.dedent(rf'''
        {{
            "file": 1,
            "format": 1
        }}{{
            "add_paused": true,
            "allow_remote": false,
            "daemon_port": {rpcport},
            "download_location": "{home}/downloads",
            "move_completed_path": "{home}/downloads",
            "listen_ports": [
                54321,
                54322
            ],
            "new_release_check": false,
            "outgoing_interface": "",
            "outgoing_ports": [
                0,
                0
            ],
            "random_outgoing_ports": false,
            "random_port": false,
            "torrentfiles_location": "{home}/torrents",
            "dht": false,
            "lsd": false,
            "natpmp": false,
            "upnp": false,
            "utpex": false
        }}
        ''').strip())
    return configdir


def run(tmp_path):
    home = common.get_home(name, tmp_path)
    configpath = get_configpath(home)
    return {
        'server_name': 'deluge',
        'server_start_cmd': f'deluged --loglevel info --do-not-daemonize --config "{configpath}"',
        'server_url': f'{username}:{password}@localhost:{rpcport}',
        'client_name': 'deluge',
    }



import asyncio

from .. import common
from . import base


class API(base.APIBase):
    def __init__(self, client):
        self.client = client

    async def perform_simple_request(self):
        result = await self.client.call('session-get')
        self.log('session-get:', result)
        assert result['result'] == 'success'
        assert result['arguments']['dht-enabled'] is False

    async def get_torrent_list(self):
        result = await self.client.call('torrent-get', fields=['hashString'])
        self.log('torrent-get:', result)
        infohashes = [torrent['hashString']
                      for torrent in result['arguments']['torrents']]
        return sorted(infohashes)

    TR_STATUS_STOPPED = 0
    TR_STATUS_CHECK_WAIT = 1
    TR_STATUS_CHECK = 2
    TR_STATUS_DOWNLOAD_WAIT = 3
    TR_STATUS_DOWNLOAD = 4
    TR_STATUS_SEED_WAIT = 5
    TR_STATUS_SEED = 6

    async def add_torrents(self, torrent_filepaths, download_path='/tmp/some/path', paused=True):
        # Add torrents
        infohashes = []
        for filepath in torrent_filepaths:
            result = await self.client.call(
                'torrent-add',
                {'download-dir': download_path},
                metainfo=common.read_torrent_file(filepath),
                paused=paused,
            )
            self.log('torrent-add:', result)
            infohashes.append(
                result['arguments']['torrent-added']['hashString'].lower()
            )

        # Wait for the server to finish verifying
        await asyncio.sleep(0.5)

        # Verify torrents where correctly added
        result = await self.client.call('torrent-get', ids=infohashes, fields=['status', 'downloadDir'])
        self.log('torrent-get:', infohashes, result)
        for torrent in result['arguments']['torrents']:
            location = torrent['downloadDir']
            status = torrent['status']
            assert location == download_path, location
            if paused:
                assert status == self.TR_STATUS_STOPPED, f'{status!r} != {self.TR_STATUS_STOPPED!r}'
            else:
                assert status == self.TR_STATUS_DOWNLOAD, f'{status!r} != {self.TR_STATUS_DOWNLOAD!r}'

        return sorted(infohashes)

    async def remove_torrents(self, infohashes):
        for infohash in infohashes:
            result = await self.client.call('torrent-remove', ids=[infohash])
            self.log(f'torrent-remove {infohash}:', result)

    async def on_torrent_added(self, handler):
        # Raise NotImplementedError
        await self.client.add_event_handler('<event name>', handler)

    async def wait_for_torrent_added(self):
        # Raise NotImplementedError
        await self.client.wait_for_event('<event name>')



class APIBase:
    def log(self, *args):
        print(f'{self.client.label} API:', ' '.join(str(arg) for arg in args))

    async def reset(self):
        # Delete all torrents
        infohashes = await self.get_torrent_list()
        if infohashes:
            self.log('Resetting: Deleting torrents:', infohashes)
            await self.remove_torrents(infohashes)
            infohashes = await self.get_torrent_list()
            assert infohashes == [], infohashes



from .. import common

name = 'transmission'

username = 'fnark'
password = 'fnorkfnork'
rpcport = '50400'


def run(tmp_path):
    home = common.get_home(name, tmp_path)
    return {
        'server_name': 'transmission',
        'server_start_cmd': (
            'transmission-daemon --log-debug --foreground '
            f'--auth --username "{username}" --password "{password}" '
            f'--config-dir "{home}/config/" '
            f'--download-dir "{home}/downloads/" '
            f'--rpc-bind-address 127.0.0.1 --port "{rpcport}" '
            '--peerport 54321 --no-portmap --no-dht --paused'
        ),
        'server_url': f'http://{username}:{password}@localhost:{rpcport}/transmission/rpc',
        'client_name': 'transmission',
    }



import textwrap

from .. import common

name = 'qbittorrent'

username = 'fnark'
password = 'fnorkfnork'
rpcport = '50200'


def get_configpath(home):
    configpath = home / 'qBittorrent/config/qBittorrent.conf'
    configpath.parent.mkdir(parents=True, exist_ok=True)
    with open(configpath, 'w') as f:
        f.write(textwrap.dedent(rf'''
        [Preferences]
        Bittorrent\DHT=false
        Connection\UPnP=false
        WebUI\Address=127.0.0.1
        WebUI\Enabled=true
        WebUI\Port={rpcport}
        WebUI\Username={username}
        WebUI\Password_PBKDF2="@ByteArray(jY/oa1dGG36Smm6tsWCKOw==:9/f0Ll63VoNHwb4b/oS/J9zskcFUPauW1i0REUY8pChwMvbIR97RKzVpftDO5wHKERCtCU3pWxXQLcL+6i/Ndw==)"

        [Network]
        PortForwardingEnabled=false

        [BitTorrent]
        Session\DHTEnabled=false
        Session\PeXEnabled=false
        Session\DefaultSavePath={home}/downloads

        [LegalNotice]
        Accepted=true
        ''').strip())
    return configpath


def run(tmp_path):
    home = common.get_home(name, tmp_path)
    get_configpath(home)
    return {
        'server_name': 'qbittorrent',
        'server_start_cmd': f'qbittorrent-nox --profile="{home}"',
        'server_url': f'{username}:{password}@localhost:{rpcport}',
        'client_name': 'qbittorrent',
    }



from .. import common


def get_common_configpath(name, tmp_path):
    configpath = common.get_home(name, tmp_path) / 'rtorrent.rc'
    logpath = common.get_home(name, tmp_path) / 'rtorrent.log'
    with open(configpath, 'w') as f:
        f.write(f'method.insert = cfg.logfile, private|const|string, (cat,"{logpath}")\n')
        f.write('print = (cat, "Logging to ",(cfg.logfile))\n')
        f.write('log.open_file = "log", (cfg.logfile)\n')
        f.write('log.add_output = "debug", "log"\n')

        # Return path to item data (never empty, unlike `d.base_path`);
        # multi-file items return a path ending with a '/'.
        f.write('method.insert = d.data_path, simple,\\\n')
        f.write('    "if=(d.is_multi_file),\\\n')
        f.write('        (cat, (d.directory)),\\\n')
        f.write('        (cat, (d.directory), /, (d.name))"\n')

        f.write('dht.mode.set = off\n')
        f.write('protocol.pex.set = no\n')
    return configpath



import asyncio
import os
import pprint

from . import base


class API(base.APIBase):
    def __init__(self, client):
        self.client = client
        self._sync_id = 0

    async def _get_sync(self):
        result = await self.client.call('sync/maindata', rid=self._sync_id)
        self._sync_id = result['rid']
        return result

    async def perform_simple_request(self):
        result = await self.client.call('app/preferences')
        self.log('app/preferences:', result)
        assert result['dht'] is False
        assert result['pex'] is False

    async def get_torrent_list(self):
        result = await self.client.call('torrents/info')
        self.log('torrents/info:', result)
        infohashes = sorted(torrent['hash'] for torrent in result)
        return infohashes

    STATUS_DOWNLOADING = 'stalledDL'
    STATUS_PAUSED = 'pausedDL'

    async def add_torrents(self, torrent_filepaths, download_path='/tmp/some/path', paused=True):
        result = await self._get_sync()
        self.log('frist sync:\n', pprint.pformat(result))

        # Add torrents
        result = await self.client.call(
            'torrents/add',
            files=[
                ('filename', (
                    os.path.abspath(filepath),
                    open(filepath, 'rb'),
                    'application/x-bittorrent',
                ))
                for filepath in torrent_filepaths
            ],
            options={'savepath': download_path, 'paused': str(paused).lower()},
        )
        self.log('torrents/add:', result)
        assert result == 'Ok.', result

        # Wait for server to finish adding, checking resume data, etc
        await asyncio.sleep(3)

        # Get added torrent hashes from server
        result = await self._get_sync()
        self.log('second sync:\n', pprint.pformat(result))
        infohashes = tuple(result['torrents'])

        # Verify torrents were correctly added
        result = await self.client.call('torrents/info', hashes='|'.join(infohashes))
        self.log('torrent/info:', infohashes, result)
        for torrent in result:
            location = torrent['save_path']
            status = torrent['state']
            assert location == download_path, location
            if paused:
                assert status == self.STATUS_PAUSED, f'{status!r} != {self.STATUS_PAUSED!r}'
            else:
                assert status == self.STATUS_DOWNLOADING, f'{status!r} != {self.STATUS_DOWNLOADING!r}'

        return sorted(infohashes)

    async def remove_torrents(self, infohashes):
        for infohash in infohashes:
            result = await self.client.call(
                'torrents/delete',
                hashes=infohash,
                deleteFiles=False,
            )
            self.log(f'torrents/delete {infohash}:', result)

    async def on_torrent_added(self, handler):
        # Raise NotImplementedError
        await self.client.add_event_handler('<event name>', handler)

    async def wait_for_torrent_added(self):
        # Raise NotImplementedError
        await self.client.wait_for_event('<event name>')
